/* UMN's wonderful Gopher Server */
/* Coded by Chris McNeil, Mount Allison University */
/* cmcneil@mta.ca  */
/* Parts of code taken from Fingerd.c  */

/*
   Modified by Peter Crawshaw (pcrawshaw@mta.ca) July 1993.
   Computing Services
   Mount Allison University
   Sackville, NB E0A 3C0

   Text file searches and dBase .dbf file searches added.
*/

/*
   Modified by Ashok Aiyar (axa12@po.cwru.edu) December 1993, January 1994.
   Dept. of Biochemistry,
   CWRU School of Medicine
   Cleveland, Ohio 44106

   Added "concatenated" index file searches that return menu items
   that match a pattern.  Imposed length restrictions on all searches.
   Additionally certain defined strings can be eliminated from
   directory searches, menu-file searches and DBF file searches.
   mkginfo function redone.  mkginfo now supports pre-made indexes too.
*/

/* Lots of bug fixes. Changed logging. Abilty to listen on more than
   one port.
   
   Selcuk Ozturk // seost2+@pitt.edu // March 31, 1995
 */

/* Copyright 1991 Phil Karn, KA9Q */

#define CSODATABASE "CSO.DBF"
#define WORDSFILE "BADWORDS.TXT"
#define DIRBAD "GOPHER.BAD"

#define GOPHER_PORT 70

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <dir.h>
#include <dos.h>
#include <ctype.h>
#include "global.h"
#include "files.h"
#include "mbuf.h"
#include "socket.h"
#include "session.h"
#include "proc.h"
#include "dirutil.h"
#include "commands.h"
#include "mailbox.h"
#include "netuser.h"
#include "ftp.h"
#include "alloc.h"
#include "smtp.h"
#ifdef GOPHER
#include "cmdparse.h"
#include "dbf.h"

#define RO_SUBDIR 0
#define RO_CASE   0
#define TRUE 1
#define FALSE 0

char *City(void);
char *Message(void);
int GophUsers;

#ifdef HTTP
/* this really should be in an *.h file shared by http.c, but .. */
struct reqInfo {
    int index;
    int method;
    int version;
    int qsize;
    int response;
    char *url;
    char *arg;
    char *query;
    char *newcheck;
    char *from;
    char *referer;
    char *agent;
    char *passwd;
    char file[128];
};
extern int authorization(int s,char *buf,struct reqInfo *rq);
#endif

/* function prototypes */
int dirmatchmake(char *pattern);
int matchmake(char *pattern);
int search(char *filename, char *pattern);
int fsplit(char *Ppart, char *Dpart, char *Fpart);
int gnamefix(int s,char *string);
int searchdirfile(int s, char *file, char *pattern);
int searchdir(int s, char *dir, char *pattern);
static void gopher __ARGS((int s,void *unused,void *p));
int split(char *input, int *type, char *file, char *pattern);
int mkginfo(int s,char *file);
int findport(int s, char *dir);
static int dogophtdisc __ARGS((int argc,char *argv[],void *p));
static int dogophsim __ARGS((int argc,char *argv[],void *p)); 
static int dogophmax __ARGS((int argc,char *argv[],void *p));
static int dogophdrive __ARGS((int argc,char *argv[],void *p));
char *decode(char *string, int isPath);


static struct cmds Gophcmds[] = {
  "drive",       dogophdrive,   0, 0, NULLCHAR,
  "maxcli",      dogophmax,     0, 0, NULLCHAR,
  "simult",      dogophsim,     0, 0, NULLCHAR, 
  "tdisc",       dogophtdisc,   0, 0, NULLCHAR,
  NULLCHAR,
};

int
dogopher(argc,argv,p)
int argc;
char *argv[];
void *p;
{
    return subcmd(Gophcmds,argc,argv,p);
}

static char DR = 'c'; 

static int
dogophdrive(argc,argv,p)
int argc;
char *argv[];
void *p;
{
  if(argc < 2)
    tprintf("Def. drive: %c\n",DR);
  else if(!isalpha(*argv[1]))
    tputs("Bad drive letter\n");
  else
    DR = *argv[1];
  return 0;
}

int Gophmax = 10;

static int
dogophmax(argc,argv,p)
int argc;
char *argv[];
void *p;
{
    return setint(&Gophmax,"Max. connects",argc,argv);
}

int Gophsimc = 5;

static int
dogophsim(argc,argv,p)
int argc;
char *argv[];
void *p;
{
    return setint(&Gophsimc,"Simult. conn.'s serviced",argc,argv);
}

int Gophtdisc = 180;

static int
dogophtdisc(argc,argv,p)
int argc;
char *argv[];
void *p;
{
    return setint(&Gophtdisc,"Server tdisc (sec)",argc,argv);
}


/* Start up gopher service */
int gstart(argc,argv,p)
int argc;
char *argv[];
void *p;
{
  int16 port;

  if(argc < 2)
    port = IPPORT_GOPHER;
  else
    port = atoi(argv[1]);

  return start_tcp(port,"Gopher Server",gopher,1024);
}


static void gopher(int s, void *unused, void *p)
{
	char command[128], *cp;
	long t;
	FILE *fp;
	char file[150],Gdir[SLINELEN],Csofile[13];
	char *dirfile;
	char *dir, *pattern;
	int length;
	char type[1];
	int i, tt,port;
	int transfer = IMAGE_TYPE; /* Transfer type, either ascii or image */


	sockmode(s, SOCK_ASCII);
	sockowner(s, Curproc);
	GophUsers++;

	if(availmem() < Memthresh || Gophmax < GophUsers) {
          close_s(s);
	  shutdown(s,2);
	  goto quit;
	}

	if(GophUsers > Gophsimc)
	  pwait(&GophUsers);

	alarm(Gophtdisc * 1000L);
	if(recvline(s, command, 128) == -1)
	  goto quit;
	alarm(0L);

	/* Find out on which local port we are */
	port = findport(s,Gdir);
	if(port == 70) {
	  strcpy(Csofile,CSODATABASE);
	} else {
	  sprintf(Csofile,"cso%u.dbf",port);
	}

	setflush(s,-1);
	rip(command);
	cp = decode(command,1);
	memcpy(command,cp,strlen(cp)+1);
	free(cp);
	length = strlen(command);
	type[0] = command[0];
	if(command[2] == ':' && command[1] != 'b' && command[1] != 'B')
		file[0] = command[1];
	else
		file[0] = DR;
	file[1] = ':';
	file[2] = '\0';

	log(s,"GOPHER #%u: %s",port,command);

	/* if blank line was received or it is a '+' */
	if((length == 0) || (!strcmp("\t+",command)) ||
	  (!strcmp("/",command)) || (!strcmp("\t$",command)) ||
	  (!strcmp("\t!",command)))
	{
		strcat(file, Gdir);
		strcat(file, "/ginfo");
	}
	else if (!strcmp(command, "1/"))
	     {
		strcat(file, Gdir);
		strcat(file, "/ginfo");
		fp = fopen(file, READ_BINARY);
		if (fp != NULLFILE)
		{
		 sendfile(fp, s, IMAGE_TYPE, 0, NULL);
		 fclose(fp);
		}
	     }
	else if (!strncmp(command, "time", 4))
	     {
		time(&t);
		usprintf(s, "\nCurrent date & time at %s:\n%s\n%s\n.\012",
			City(), ptime(&t), Message());
		goto quit;
	     }
	else if (!strncmp(command, "fields", 6))
	     {
		transfer = IMAGE_TYPE;
		sockmode(s, SOCK_BINARY);
		dbf_fields(s, Csofile, CSOTEXT);
		usprintf(s, "200:Ok.\n");
		type[0] = '2';
	     }
	else if (!strncmp(command, "query", 5) || !strncmp(command,"ph",2))
	     {
		transfer = IMAGE_TYPE;
		sockmode(s, SOCK_BINARY);
		for (i=0; i < strlen(command); i++)
		   if (command[i] == '"')
		       command[i] = ' ';
		cp = skipnonwhite(command);
		dbf_do_search(s,++cp, CSOTEXT,Csofile);
		type[0] = '2';
	     }
	else {
	/* copy command to file name, delete any .. to avoid */
	/* anyone telnetting to gopher and cd'ing to a directory */
	/* other than drive:\gopher\
					*/
	  { int k;

	     if(command[2] == ':') {
		k = 3;
	      } else {
		k = 1;
	      }
		strcat(file, Gdir);
		for ( i=k; i <= strlen(command); i++ ) {
		  if (command[i] == '.' && command[i+1] == '.') {
		    command[i] = ' ';
		    command[i+1] = ' ';
		  }
		  file[(strlen(Gdir)+i+2-k)] = command[i];
		}
	  }
	}
#ifdef HTTP        /* used only if HTTP server is also compiled in */
        /* Let's check the access file in /wwwstat dir. It is primarily
            for http but we check here, too in case, the http & gopher
            dirs coincide. They do by default. */
        {
            struct reqInfo rq;
            char buf[256];
            
            memset(&rq,0,sizeof(rq));
            rq.url = &file[strlen(Gdir)+3];
            /* authorization() will send a html message body, but that's ok */
            if(!authorization(s,buf,&rq))
                goto quit;
        }
#endif
            
	switch (type[0]) {
		case '1':
			transfer = IMAGE_TYPE;
			strcat(file, "/ginfo");
			break;
		case '2':
		case '3':
			transfer = ASCII_TYPE;
			break;
		case '0':
		case '4':
		case '5':
		case '6':
		case '8':
		case '9':
			transfer = IMAGE_TYPE;
			break;
		case '7':
			transfer = ASCII_TYPE;
			dir = strtok(file, "\t");
			pattern = strtok(NULL, "\t");
			searchdir(s, dir, pattern);
			break;
		case 's':
			transfer = ASCII_TYPE;
			dirfile = strtok(file,"\t");
			pattern = strtok(NULL, "\t");
			searchdirfile(s, dirfile, pattern);
			break;
		/* send a listing of the fields in a dbf file */
		case 'p':
			transfer = ASCII_TYPE;
			dbf_fields(s, file, ASCIITEXT);
			break;
		/* send a back modified selector string for a dbf search */
		case 'q':
			transfer = ASCII_TYPE;
			for (i=0; i < strlen(file); i++)
			{
			  if (file[i] == '\t')
			    file[i] = '$';
			}
			usprintf(s, "0Results of your search\tz%s\t%s\t%u\n",
					   file, Hostname,port);
			break;
		/* do a search on a dbf file and send the results to the client */
		case 'z':
			transfer = ASCII_TYPE;
			dbf_do_search(s, file, ASCIITEXT,"");
			break;
		default:
			transfer = IMAGE_TYPE;
			break;
		}

	switch(type[0])
	{
	   case 's':
	   case 'p':      /* Don't send a file for search types */
	   case 'q':
	   case 'z':
	   case '2':
	   case '7':
	      break;
	   default:

	      if (transfer == ASCII_TYPE)
		fp = fopen(file, READ_TEXT);
	      else
		fp = fopen(file, READ_BINARY);

	      if (fp != NULLFILE)
		{
		  sendfile(fp, s, transfer, 0, NULL);
		  fclose(fp);
		}
	      else
		/* If no ginfo file lets make one on the fly */
		if (type[0] == '1' ) {
		  mkginfo(s,file);
		}
		else usprintf(s, "ERROR: Could not find requested file");
	      
	    }    /* switch */
	/* If text files don't end with a blank line */

	if ( (transfer==ASCII_TYPE) && ( (type[0] != '0') && (type[0] != '2') ) )
	       usprintf(s, ".\012");
 
quit:
	usflush(s);
	close_s(s);
	GophUsers--;
	psignal(&GophUsers,1); 
}


int g0(int argc, char *argv[], void *p)
{
  int16 port;

  if(argc < 2)
    port = IPPORT_GOPHER;
  else
    port = atoi(argv[1]);
  
  return stop_tcp(port);
}

char *Message()
{
	static char str[100];
	char *p, *getenv();

		if ((p = getenv("MESSAGE")) == NULL)
			sprintf(str,"Have a nice day!");
		else
			sprintf(str,"%s", p);
	return(str);
}

char *City()
{
	static char str[50];
	char *p, *getenv();

		if ((p = getenv("CITY")) == NULL)
			sprintf(str,"this location");
		else
			sprintf(str,"%s", p);
	return(str);
}

int searchdir(int s, char *dir, char *pattern)
/*
   Search a directory for files containig 'pattern'.
   Output the results on socket s.
*/
{
    struct ffblk SR;
    int done,port;
    static int result;
    char buf[100], Dpart[67], Npart[13], Gdir[SLINELEN];
    char *p, tmp[13];
    
    port = findport(s,Gdir);
    Dpart[0] = '\0';
    Npart[0] = '\0';
    undosify(dir);
    fsplit(dir, Dpart, Npart);
    if (strlen(Npart) == 0)
      strcpy(Npart, "*.*");

    strcpy(buf, dir);
    if (buf[strlen(buf)-1] == '/')
      strcat(buf, Npart);

    if (strlen(pattern) <3)
    {
      gnamefix(s,buf);
      usprintf(s,"0'%s' is too short (minimum search string is 3 characters)\t0bogus\tbogus.host\t70\n7Repeat search with a longer search string\t7%s\t%s\t%u\n", pattern, buf, Hostname,port);
      return 0;
    }
    else
    if ((matchmake(pattern)) == 1)
	{
	  gnamefix(s,buf);
	  usprintf(s,"0'%s' is not a permitted search string\t0bogus\tbogus.host\t70\n7Repeat search with another search string\t7%s\t%s\t%u\n", pattern, buf, Hostname,port);
	  return 0;
	}
    else
    done = findfirst(buf, &SR, FA_ARCH);
    while (!done)
    {
      strcpy(buf, Dpart);
      strcat(buf, SR.ff_name);
      undosify(buf);
	if (((result = search(buf, pattern))==1))
	{
	gnamefix(s,buf);
	p = strrchr(buf, '/');
	strcpy(tmp, ++p);
	usprintf(s, "0%s\t0%s\t%s\t%u\n", p, buf, Hostname,port);
	}
      pwait(NULL);
      done = findnext(&SR);
      }

    return result;
}


int matchmake(char *pattern)
{
  FILE *WorkFile;
  char Line[80];
  int result;

  if((WorkFile = fopen(WORDSFILE, "rt")) != NULLFILE){
    while (!feof(WorkFile))
    {
    fgets(Line, 80, WorkFile);
      strupr(Line);
      strupr(pattern);
    if (strstr(Line, pattern) != NULL)
      {
	result = 1;
	break;
      }
    }
  fclose(WorkFile);
  }
  else result = 0;
  return(result);
}

int searchdirfile(int s, char *file, char *pattern)
/*
  Search a file for lines that match a particular pattern
  Return the result on socket s
*/
{
    FILE *f;
    char *p, Line[240], upperLine[240],Gdir[SLINELEN];
    int result,port;

    port = findport(s,Gdir);
    strupr(pattern);

    if (strlen(pattern) <3)
    {
      gnamefix(s,file);
      usprintf(s,"0'%s' is too short (minimum search string is 3 characters)\t0bogus\tbogus.host\t70\n7Repeat search with a longer search string\ts%s\t%s\t%u\n", pattern, file, Hostname,port);
      return 0;
    }
    else
    if ((dirmatchmake(pattern)) == 1)
	{
	  gnamefix(s,file);
	  usprintf(s,"0'%s' is not a permitted search string\t0bogus\tbogus.host\t70\n7Repeat search with another search string\ts%s\t%s\t%u\n", pattern, file, Hostname,port);
	  return 0;
	}
    else
    if ((f = fopen(file, "rt")) != NULLFILE)  {
	while (!(feof(f)))  {
	  fgets(Line, 240, f);
	  strncpy(upperLine, Line, 240);
	  p = strtok(upperLine, "\t");
	  strupr(p);
	  if (strstr(p, pattern) != NULL) {
	    usprintf(s, "%s", Line);
	    result = 1;
	    continue;
	  }
	  pwait(NULL);
	}
  fclose(f);
  }
  else result = 0;
  return(result);
}


int dirmatchmake(char *pattern)
{
  FILE *WorkFile;
  char Line[80];
  int result;

  if((WorkFile = fopen(DIRBAD, "rt")) != NULLFILE){
    while (!feof(WorkFile))
    {
    fgets(Line, 80, WorkFile);
    strupr(pattern);
    strupr(Line);
    if (strstr(Line, pattern) != NULL)
      {
	result = 1;
	break;
      }
    }
  fclose(WorkFile);
  }
  else result = 0;
  return(result);
}


int search(char *filename, char *pattern)
/*
  Do text file searches for substrings.
  Returns TRUE if the file contains 'pattern', false otherwise.
*/
{
  FILE *TextFile;
  char Line[80];
  int result;

  if((TextFile = fopen(filename, "rt")) != NULLFILE){
  while (!feof(TextFile))
  {
    fgets(Line, 80, TextFile);
    strupr(Line);
    strupr(pattern);
    if (strstr(Line, pattern) != NULL)
    {
      result = 1;
      break;
    }
    pwait(NULL);
  }
  fclose(TextFile);
  }
  else result = 0;
  return(result);
}


int fsplit(char *Ppart, char *Dpart, char *Fpart)
/*
  Split a file path into its directory and filename parts.
*/
{
  char *p;

  strcpy(Dpart, Ppart);
  p = strrchr(Dpart, '/');
  if (!p)
    p = strrchr(Dpart, '/');
  if (p)
  {
     p++;
     strcpy(Fpart, p);
     *p = '\0';
  }
  return 0;
}


int gnamefix(int s,char *string)
/*
  Get rid of '/gopher' from a file path string.
*/
{
  char *p;
  char tmp[128],Gdir[SLINELEN];

  findport(s,Gdir);

  strlwr(string);
  if ((p = strstr(string,Gdir)) != NULLCHAR)
  {
     *p = '\0';
     p += strlen(Gdir);
     strcpy(tmp, string);
     strcat(tmp, p);
     strcpy(string, tmp);
  }
  return 0;
}

#ifdef MAILLOG
int mail_logfile(char *user, char *host)
{
	struct sockaddr_in sock;
	char *cp;
	int s,i, reply;
	struct mbuf *bp;
	FILE *f;
	char buf[512];

	sock.sin_family = AF_INET;
	sock.sin_port = IPPORT_SMTP;

	if((sock.sin_addr.s_addr = resolve(host)) == 0)
	{
		tprintf("Gopher Log Mail Host %s unknown\n", host);
		return -1;
	}

	if((s = socket(AF_INET,SOCK_STREAM,0)) == -1)
	{
		tprintf("Can't create socket\n");
		return -1;
	}
	sockmode(s,SOCK_ASCII);
	if(connect(s,(char *)&sock,sizeof(sock)) == -1)
	{
		cp = sockerr(s);
		tprintf(" -- Connect failed: %s\n",cp != NULLCHAR ? cp : "");
		close_s(s);
		return -1;
	}

	recvline(s, buf, sizeof(buf));

	sprintf(buf, "HELO %s\n", Hostname);
	usprintf(s, "%s", buf);

	recvline(s, buf, sizeof(buf));
	sscanf(buf, "%d", &reply);
	if (reply != 250)
	{
	   tprintf("%s\n", buf);
	   return -1;
	}

	sprintf(buf, "MAIL FROM: <gopher@%s>\n", Hostname);
	usprintf(s, "%s", buf);

	recvline(s, buf, sizeof(buf));
	sscanf(buf, "%d", &reply);
	if (reply != 250)
	{
	   tprintf("%s\n", buf);
	   return -1;
	}

	sprintf(buf, "RCPT TO: <%s@%s>\n", user, host);
	usprintf(s, "%s", buf);

	recvline(s, buf, sizeof(buf));
	sscanf(buf, "%d", &reply);
	if (reply != 250)
	{
	   tprintf("%s\n", buf);
	   return -1;
	}

	sprintf(buf, "DATA\n");
	usprintf(s, "%s", buf);

	recvline(s, buf, sizeof(buf));
	sscanf(buf, "%d", &reply);
	if (reply != 354)
	{
	   tprintf("%s\n", buf);
	   return -1;
	}

	sprintf(buf, "To: %s@%s\n", user, host);
	usprintf(s, "%s", buf);
	sprintf(buf, "From: gopher@%s\n", Hostname);
	usprintf(s, "%s", buf);

	f = fopen(GopherLog, READ_TEXT);
	while (!feof(f))
	{
	    fgets(buf, sizeof(buf), f);
	    usprintf(s, "%s", buf);
	    pwait(NULL);
	}
	fclose(f);

	usprintf(s, "\n.\n");
	sprintf(buf, "BYE\n");
	usprintf(s, "%s", buf);

	close_s(s);

	remove(GopherLog);
	f = fopen(GopherLog, APPEND_TEXT);
	fclose(f);

	return 0;
}


int sendgopherlog (int argc, char *argv[], void *p)
{
    char *pp, *user, *host;

    if (argc != 2)
    {
      tprintf("Invalid number of args %d\n", argc);
      return -1;
    }

    user = argv[1];
    if((pp = strrchr(argv[1], '@')) == NULLCHAR) {
      host = Hostname;      /* send it to ourselves.  Selcuk */
    } else {
      *pp = '\0';
      pp++;
      host = pp;
    }
    mail_logfile(user, host);

    return 0;
}
#endif /* ifdef MAILLOG */

int mkginfo(int s, char *file)
{
	FILE *fp;
	char dirstring[128];
	char file1[140];
	char *p,*ext;
	int length;
	int i, tt, done, doit, port;
	char type[2];
	char transfer[2]; /* Transfer type, either ascii or image */
	struct ffblk ffblk;
	char temp[128],Gdir[SLINELEN];
	char drive[3];


	drive[0]=file[0];
	drive[1]=file[1];
	drive[2]='\0';
	type[1]='\0';
	transfer[1]='\0';
	file1[0]='\0';

	port = findport(s,Gdir);
	strrev(file);
	strcpy(dirstring,strstr(file,"/"));
	strrev(dirstring);

	done=findfirst(strcat(dirstring,"*.*"),&ffblk,FA_DIREC);

	if ( !done ) {
	strtok(dirstring,"*");
	/* get rid of /gopher part */
	p = dirstring;
	p += strlen(Gdir) + 2;
	undosify(p);
	}

		while (!done) {
		doit=1;
		/* Get the file extension */
		ext=(strstr(ffblk.ff_name,"."));
		if ( (!strcmpi(ext,".zip")) || (!strcmpi(ext,".arc")) || (!strcmpi(ext,".arj"))
		    || (!strcmpi(ext,".lha")) || (!strcmpi(ext,".zoo")) )
		 {
		 type[0]='5';           /* DOS binary archive only */
		 transfer[0]='9';
		 }
		 else if ( (!strcmpi(ext,".exe")) || (!strcmpi(ext,".com")) || (!strcmpi(ext,".ovr"))
			|| (!strcmpi(ext,".dll")) || (!strcmpi(ext,".wp"))  || (!strcmpi(ext,".z"))
			|| (!strcmpi(ext,".wb1")) || (!strcmpi(ext,".msw")) || (!strcmpi(ext,".taz"))
			|| (!strcmpi(ext,".hlp")) || (!strcmpi(ext,".ovl")) || (!strcmpi(ext,".tgz"))
			|| (!strcmpi(ext,".tar")) || (!strcmpi(ext,".wri")) || (!strcmpi(ext,".doc"))
			|| (!strcmpi(ext,".wk1")) || (!strcmpi(ext,".wrk")) || (!strcmpi(ext,".dvi"))
			|| (!strcmpi(ext,".wp5")) || (!strcmpi(ext,".wp6")) )
		 {
		 type[0]='9';           /* All other binary file types */
		 transfer[0]='9';
		 }
		 else if (!strcmpi(ext,".hqx"))
		 {
		 type[0]='4';           /* Mac BinHex file */
		 transfer[0]='0';
		 }
		 else if (!strcmpi(ext,".uue"))
		 {
		 type[0]='6';           /* Unix UUencoded file */
		 transfer[0]='0';
		 }
		 else if ( (!strcmpi(ext,".bmp")) || (!strcmpi(ext,".jpg")) || (!strcmpi(ext,".icn"))
			|| (!strcmpi(ext,".pcx")) || (!strcmpi(ext,".xbm")) || (!strcmpi(ext,".pic"))
			|| (!strcmpi(ext,".wpg")) || (!strcmpi(ext,".tif")) || (!strcmpi(ext,".pbm"))
			|| (!strcmpi(ext,".jif")) || (!strcmpi(ext,".gif")) || (!strcmpi(ext,".ico"))
			|| (!strcmpi(ext,".eps")) || (!strcmpi(ext,".rle")) )
		 {
		 type[0]='I';           /* All image files */
		 transfer[0]='9';
		 }
		 else if ( (!strcmpi(ext,".qt")) || (!strcmpi(ext,".dl")) || (!strcmpi(ext,".gl"))
			|| (!strcmpi(ext,".mpg")) || (!strcmpi(ext,".fli")) || (!strcmpi(ext,".qtw")) )
		 {
		 type[0]=';';           /* movies */
		 transfer[0]='9';
		 }
		 else if ( (!strcmpi(ext,".au")) || (!strcmpi(ext,".wav"))
			|| (!strcmpi(ext,".snd")) || (!strcmpi(ext,".sou")) )
		 {
		 type[0]='s';           /* Sound files */
		 transfer[0]='9';
		 }
		 else if (!strcmpi(ext,".dbf"))
		 {
		 type[0]='1';           /* database files */
		 transfer[0]='p';
		 }
		 else if (!strcmpi(ext,".htm"))
		 {
		 type[0]='h';
		 transfer[0]='0';
		 }
/*               else if ( (!strcmpi(ext,".cat")) || (!strcmpi(ext,".ftp")) )
		 {
		 doit=0;
		 } */
		 else if ( ffblk.ff_attrib == FA_DIREC )
		 {
			 if (ffblk.ff_name[0] == '.')
			 {
			 doit=0;
			 }
			 else
			 {
			 type[0]='1';
			 transfer[0]='1';
			 }
		 }
		 else
		 {
		type[0]='0';
		transfer[0]='0';
		}
		if (doit==1)
		{
		if (transfer[0]=='p') usprintf(s, "%sSearch %s\t%s%s%s%s\t%s\t%u\012",type, ffblk.ff_name,transfer,drive,p,ffblk.ff_name, Hostname,port);
		else if (transfer[0]=='1') usprintf(s, "%s%s\t%s%s%s%s\t%s\t%u\012",type, strlwr(ffblk.ff_name),transfer,drive,p,ffblk.ff_name, Hostname,port);
		else usprintf(s, "%s%-17s[%ld bytes  %d-%d-%d]\t%s%s%s%s\t%s\t%u\012",type, strlwr(ffblk.ff_name),ffblk.ff_fsize,
		(ffblk.ff_fdate >> 5) & 0xf,/* month */
		(ffblk.ff_fdate ) & 0x1f,/* day */
		(ffblk.ff_fdate >> 9) + 80/* year */
		,transfer,drive,p,ffblk.ff_name, Hostname,port);
		}
		pwait(NULL);
		done=findnext(&ffblk);
	     }
	strcat(file1, strcat(dirstring,"menu.cat"));
	fp = fopen(file1, READ_BINARY);
	if (fp != NULLFILE)
	{
	sendfile(fp, s, 0, IMAGE_TYPE, NULL);
	fclose(fp);
	}
	return 0;
}

int findport(int s, char *dir)
{
  int i;
  struct sockaddr_in lsocket;

  i = SOCKSIZE;
  getsockname(s,(char *)&lsocket,&i);


  if(lsocket.sin_port == 70)
    strcpy(dir,Gopherdir);
  else 
    sprintf(dir,"/g%u",lsocket.sin_port);

  return lsocket.sin_port;
}
#endif /* GOPHER */

#if defined(HTTP) || defined(GOPHER)
/* Creates a copy of string which has '%' escaped charecters decoded
   back to ASCII and '+' signs to ' ' & '&' to 0xff if isPath == 0.
   If isPath == 1, ie. decoding a path not a query string, then
   it is canonized so that authorization can identify the path later.
   
   Allocates memory. Caller must free it. 
*/

char *decode (char *string, int isPath)
{
int i=0,k=0,slen;
char code[3],*tmp;

	slen = min(strlen (string),99);
        tmp = (char *)mallocw(slen+1);
        
	code[2] = '\0';

        /* possible security breach */
            
	while (i <= slen) {
            if (string[i] == '%') {
/*                string[i] = '#';        /* '%' sign causes problems in log()  */
		code[0] = string[i+1];
		code[1] = string[i+2];
		tmp[k] = (char)htoi(code);
		k++;  i += 3;
            } else {
                tmp[k] = string[i];
                if(!isPath) {
                    if(tmp[k] == '+') {
                        tmp[k] = ' ';
                    } else if(tmp[k] == '&') {
                        tmp[k] = '\xff';
                    }
                }
                k++; i++;
            }
	}
        
        /* now, cononize the path; DOS accepts different strings for the
           same path and that causes hell in authorization */
        if(isPath) {  // we might break something for the gopher here, but I don't think so
            undosify(tmp);
            if(*tmp == '.' || *tmp == '/')
                *tmp = 0;           // they'll get the root.htm
            
            i = 0; k = 0; slen = strlen(tmp);
            while(i <= slen) {
                if((tmp[i] == '.' || tmp[i] == '/') &&
                      (tmp[i+1] == '.' || tmp[i+1] == '/')) {
                    i++;
                } else {
                    tmp[k++] = tmp[i++];
                }
            }
        }
	return tmp;
}

#endif
